from __future__ import annotations

import typing as t
from textwrap import indent
from time import localtime
from time import strftime as _strftime

from ._internal_debug import debug  # noqa

strftime = lambda t: _strftime('%H:%M:%S', localtime(t)) \
    if t is not None else ''


class Color:  # TODO: not used yet (and not ready).
    SEPARATOR = 'bright_black'
    
    # time
    TIME_START = 'green'
    DURATION_S = 'green'
    DURATION_M = 'green_yellow'
    DURATION_L = 'yellow'
    DURATION_XL = 'red'
    
    # verbosity
    V_DEBUG = 'grey50'
    V_INFO = 'blue'
    V_WARNING = 'yellow'
    V_ERROR = 'red'
    V_FATAL = 'bold #ffffff on red'


class MessageFormatter:
    
    @staticmethod
    def markup(*markups: tuple[str, str]) -> str:
        """
        this method produces the final strings that could be directly printed.

        args:
            markups: tuple[tuple[str text, str mark], ...]
                mark:
                    any valid patterns of pytermgui markups. for example, 'red',
                    'bold', 'dim', '#00FF00', ...
                    ps: mark allows empty.
        """
        out = []
        for text, mark in markups:
            if mark:
                out.append('[{}]{}[/]'.format(mark, text))
            else:
                out.append(text)
        return ''.join(out)
    
    # -------------------------------------------------------------------------
    
    def fmt_source(self, filepath: str, lineno: t.Union[int, str],
                   is_external_lib: bool = False, fmt_width=False) -> str:
        if fmt_width:
            text_a = f'{filepath}:{lineno}'
            text_b = self._fmt_width(text_a, min_width=12)
            additional_space = ' ' * (len(text_b) - len(text_a))
        else:
            additional_space = ''
        if is_external_lib:
            assert filepath.startswith('[')
            a, b = filepath[1:].split(']', 1)
            filepath = self.markup(
                ('\\[{}]'.format(a),
                 'bold {}'.format('red' if filepath.startswith('[unknown]')
                                  else 'magenta')),
                (f'{b}', 'bold blue'),
            )
            return self.markup(
                (filepath, ''),
                (':', 'bright_black'),
                (str(lineno), 'bold blue'),
                (additional_space, ''),
            )
        else:
            return self.markup(
                (filepath.replace('[', '\\['), 'bold blue'),
                (':', 'bright_black'),
                (str(lineno), 'bold blue'),
                (additional_space, ''),
            )
    
    def fmt_separator(self, sep: str = ' >> ', color='bright_black') -> str:
        return self.markup((sep, color))
    
    def fmt_funcname(self, funcname: str, fmt_width=False) -> str:
        is_func = not funcname.startswith('<')
        if is_func:
            funcname += '()'
        else:
            funcname = funcname[1:-1]
        if fmt_width:
            funcname = self._fmt_width(
                funcname, min_width=8, unit_spaces=4
            )
        if is_func:
            return self.markup((funcname, 'green'))
        else:
            return self.markup((funcname, 'magenta'))
    
    def fmt_index(self, idx: int) -> str:
        return self.markup(
            (f'[{idx}]', 'grey50' if idx == 0 else 'red')
        )
    
    @staticmethod
    def fmt_time(start: float, end: float = None, color_s='green') -> str:
        if end is None:
            return '[{}]\\[{}][/]'.format(color_s, strftime(start))
        
        diff = end - start  # float >= 0
        
        color_s: str
        color_e: str
        color_d: str
        
        if diff < 0.500:
            color_e = 'green'
            color_d = 'green'
        elif diff < 1.000:
            color_e = 'green'
            color_d = 'green_yellow'
        elif diff < 5.000:
            color_e = 'yellow'
            color_d = 'yellow'
        else:
            color_e = 'red'
            color_d = 'red'
        if int(start) == int(end):
            color_e += ' dim'
            color_d += ' dim'
        
        # better diff
        if diff < 1.000:
            diff = '{:.0f}ms'.format(diff * 1000)
        else:
            diff = '{:.1f}s'.format(diff)
        
        return '[{color_s}]\\[{start}][/]' \
               '[grey50] -> [/]' \
               '[{color_e}]\\[{end}][/] ' \
               '[{color_d}]({diff:>5})[/]'.format(
            start=strftime(start),
            end=strftime(end),
            diff=diff,
            color_s=color_s,
            color_e=color_e,
            color_d=color_d,
        )
    
    def fmt_divider(self, div_: str = '-' * 64) -> str:
        return self.markup((div_, 'yellow'))
    
    def fmt_message(self, arguments: t.Iterable[str], varnames: tuple[str],
                    rich: bool, expand=False, separator=';   ') -> str:
        if varnames:
            arguments = self._mix_arguments_with_varnames(arguments, varnames)
        if not rich:
            arguments = (x.replace('[', '\\[') for x in arguments)
        if expand:  # FIXME
            return '\n' + indent('\n'.join(arguments), '    ')
            # from rich.pretty import pretty_repr
            # arguments = (pretty_repr(x) for x in arguments)
            # return '\n' + indent('\n'.join(arguments), '    ')
        else:
            return self.markup((separator, 'bright_black')).join(arguments)
    
    def fmt_level(self, level: str, custom_text='') -> str:
        if level == 'trace':
            return ''
        labels = {
            'trace': '',
            'debug': '[DEBUG]',
            'info' : ' [INFO]',
            'warn' : ' [WARN]',
            'error': '[ERROR]',
            'fatal': '[FATAL]',
        }
        colors = {
            'trace': '',
            'debug': 'grey50',
            'info' : 'blue',
            'warn' : 'yellow',
            'error': 'red',
            'fatal': 'bold #ffffff on red',
        }
        if custom_text:
            text = custom_text
        else:
            text = labels[level]
        return self.markup((text, colors[level]))
    
    # -------------------------------------------------------------------------
    
    @staticmethod
    def _mix_arguments_with_varnames(
            arguments: t.Iterable[str], varnames: tuple[str, ...],
    ) -> t.Iterable[str]:
        if not arguments:
            return ()
        try:
            # noinspection PyTypeChecker
            assert len(varnames) == len(arguments), (varnames, arguments)
        except AssertionError:
            # debug('failed extracting varnames')
            return arguments
        else:
            return tuple(f'{v} = {a}' if v else str(a)
                         for v, a in zip(varnames, arguments))
    
    @staticmethod
    def _fmt_width(text: str, min_width: int = None, unit_spaces=4) -> str:
        if len(text) <= min_width:
            return text.ljust(min_width)
        else:
            from math import ceil
            return text.ljust(ceil(len(text) / unit_spaces) * unit_spaces)



from __future__ import annotations

from inspect import currentframe

from ._internal_debug import debug  # noqa
from .console import con_print

__all__ = ['LKLogger', 'lk']


class T:  # Typehint
    from rich.console import RenderableType
    from typing import Iterable, TypedDict, Union
    from .markup import T as _TMarkup  # noqa
    
    Args = Union[list[str], tuple[str, ...]]
    MarkupPos = int
    Markup = str
    Marks = _TMarkup.Marks
    MarksMeaning = _TMarkup.MarksMeaning
    
    Info = TypedDict('Info', {
        'file_path'      : str,
        'line_number'    : str,
        'is_external_lib': bool,
        'function_name'  : str,
        'variable_names' : Iterable[str],
    })


class LKLogger:
    
    def __init__(self):
        from .cache import LoggingCache
        from .config import LoggingConfig
        from .markup import MarkupAnalyser
        from .message_builder import MessageBuilder
        
        self._analyser = MarkupAnalyser()
        self._builder = MessageBuilder()
        self._cache = LoggingCache()
        self._config = LoggingConfig()
    
    def configure(self, clear_preset=False, **kwargs) -> None:
        self._cache.clear_cache()
        if clear_preset:
            self._config.reset()
        self._config.update(**kwargs)
        self._builder.update_config(
            separator=self._config.separator,
            show_source=self._config.show_source,
            show_funcname=self._config.show_funcname,
            show_varnames=self._config.show_varnames,
        )
    
    # -------------------------------------------------------------------------
    
    def log(self, *args, **kwargs) -> None:
        msg = self._build_message(currentframe().f_back, *args)
        # debug(msg)
        con_print(msg, **kwargs)
    
    def fmt(self, *args, **_) -> str:
        return str(self._build_message(currentframe().f_back, *args))
    
    def _build_message(self, frame, *args) -> T.RenderableType:
        from .markup import MarkMeaning
        
        frame_id = f'{id(frame)}#{frame.f_lineno}'
        args, markup_pos, markup = \
            self._extract_markup_from_arguments(frame_id, args)
        marks = self._analyser.extract(markup)
        marks_meaning = self._analyser.analyse(marks)
        
        # check cache
        if self._cache.is_cached(frame_id, markup):
            cached_info = self._cache.get_cache(frame_id, markup)
            return self._builder.compose(
                args, marks_meaning, cached_info
            )
        
        # ---------------------------------------------------------------------
        
        if MarkMeaning.AGRESSIVE_PRUNE in marks_meaning:
            return self._builder.quick_compose(args)
        
        info: T.Info = {
            'file_path'      : '',
            'line_number'    : '0',
            'is_external_lib': False,
            'function_name'  : '',
            'variable_names' : (),
        }
        
        show_source = self._config.show_source
        show_funcname = self._config.show_funcname
        show_varnames = self._config.show_varnames and \
                        MarkMeaning.MODERATE_PRUNE not in marks_meaning
        
        if any((show_source, show_funcname, show_varnames)):
            from .sourcemap import sourcemap
            srcmap = sourcemap.get_sourcemap(
                frame=frame,
                traceback_level=marks['p'],
                advanced=show_varnames,
            )
            
            if show_source:
                
                def update_sourcemap():
                    """
                    this function updates follows:
                        info['is_external_lib']
                        info['file_path']
                        info['line_number']
                    """
                    from .path_helper import path_helper
                    path = srcmap.filepath
                    info['is_external_lib'] = path_helper.is_external_lib(path)
                    
                    if info['is_external_lib']:
                        if self._config.show_external_lib:
                            style = self._config.path_style_for_external_lib
                            info['file_path'] = \
                                path_helper.reformat_external_lib_path(
                                    path, style
                                )
                        else:
                            info['file_path'] = ''
                    else:
                        info['file_path'] = path_helper.relpath(path)
                    
                    info['line_number'] = str(srcmap.lineno)
                
                update_sourcemap()
            
            if show_funcname:
                info['function_name'] = srcmap.funcname
            
            if show_varnames:
                if markup_pos == 0:
                    info['variable_names'] = srcmap.varnames
                elif markup_pos == 1:
                    info['variable_names'] = srcmap.varnames[1:]
                else:
                    info['variable_names'] = srcmap.varnames[:-1]
        
        self._cache.store_info(frame_id, markup, info)
        
        return self._builder.compose(args, marks_meaning, info)
    
    def _extract_markup_from_arguments(
            self, frame_id: str, args: T.Args
    ) -> tuple[T.Args, T.MarkupPos, T.Markup]:
        if (markup_pos := self._cache.get_markup_pos(frame_id)) is None:
            is_markup = self._analyser.is_valid_markup
            if (
                    len(args) > 0 and
                    isinstance(args[0], str) and
                    args[0].startswith(':') and
                    is_markup(args[0])
            ):
                markup_pos = 1
            elif (
                    len(args) > 1 and
                    isinstance(args[-1], str) and
                    args[-1].startswith(':') and
                    is_markup(args[-1])
            ):
                markup_pos = -1
            else:
                markup_pos = 0
            self._cache.record_markup_pos(frame_id, markup_pos)
        
        # markup_pos: 0 not exists, 1 first place, -1 last place.
        args = tuple(map(str, args))
        if markup_pos == 0:
            return args, markup_pos, ''
        if markup_pos == 1:
            return args[1:], markup_pos, args[0]
        else:
            return args[:-1], markup_pos, args[-1]


lk = LKLogger()



from .general import default_print
from .global_control import disable
from .global_control import enable
from .global_control import setup
from .global_control import start_ipython
from .global_control import unload
from .global_control import update
from .logger import lk

__version__ = '5.2.3'



"""
what we need:
    1. if source came from '~/libs/site-packages', show short form of path.
    2. use built-in print method for external source paths.
    3. do not print anything for external source paths.
    4. ...
"""


class PathFilter:
    
    def __init__(self):
        self.whitelist = set()
        self.blacklist = set()
        self.policy = None
    
    def block_site_packages(self):
        pass
    
    def block_any_externals(self):
        pass
    
    def filter_dir(self, dirpath: str):
        pass
    
    def check_pass(self, path: str) -> bool:
        pass



from __future__ import annotations

import typing as t
from enum import Enum
from enum import auto
from time import time


# noinspection PyArgumentList
class MarkMeaning(Enum):
    AGRESSIVE_PRUNE = auto()
    DIVIDER_LINE = auto()
    EXPAND_MULTIPLE_LINES = auto()
    MODERATE_PRUNE = auto()
    RESET_INDEX = auto()
    RESET_TIMER = auto()
    RICH_FORMAT = auto()
    START_TIMER = auto()
    STOP_TIMER = auto()
    UPDATE_INDEX = auto()
    VERBOSITY = auto()


class T:
    Markup = str
    Marks = t.TypedDict('Marks', {
        'd': int,  # divider line
        'i': int,  # index
        'l': int,  # long / loose format (multiple lines)
        'p': int,  # parent layer
        'r': int,  # rich
        's': int,  # short / simple / single line
        't': int,  # timestamp
        'v': int,  # verbose / log level
    })
    MarksMeaning = dict[MarkMeaning, t.Any]


class MarkupAnalyser:
    """
    readme: prj:/docs/markup.zh.md
    """
    from re import compile
    _mark_pattern_0 = compile(r'^:(?:[dilprstv][0-9]?)+$')
    _mark_pattern_1 = compile(r'\w\d?')
    
    def is_valid_markup(self, text: str) -> bool:
        return bool(self._mark_pattern_0.match(text))
    
    def extract(self, markup: T.Markup) -> T.Marks:
        """
        return:
            dict[literal mark, int value]
        """
        out = {
            'd': -1,
            'i': -1,
            'l': -1,
            'p': 0,
            'r': -1,
            's': -1,
            't': -1,
            'v': 0,
        }
        defaults = {
            'd': 0,
            'i': 1,
            'l': 0,
            'p': 1,
            'r': 0,
            's': 0,
            't': 2,
            'v': 1,
        }
        for m in (self._mark_pattern_1.findall(markup) or ()):
            if len(m) == 1:
                out[m[0]] = defaults[m[0]]
            else:
                out[m[0]] = int(m[1:])
        return out
    
    _simple_count = 0
    _simple_time = time()
    
    def analyse(self, marks: T.Marks) -> T.MarksMeaning:
        out = {}
        
        if marks['d'] >= 0:
            out[MarkMeaning.DIVIDER_LINE] = '-' * 64
        
        if marks['i'] == 0:
            self._simple_count = 0
            out[MarkMeaning.RESET_INDEX] = 0
            out[MarkMeaning.RICH_FORMAT] = True
        elif marks['i'] >= 1:
            self._simple_count += 1
            out[MarkMeaning.UPDATE_INDEX] = self._simple_count
        
        if marks['l'] >= 0:
            out[MarkMeaning.EXPAND_MULTIPLE_LINES] = True
        
        if marks['r'] >= 0:
            out[MarkMeaning.RICH_FORMAT] = True
        
        if marks['s'] == 0:
            out[MarkMeaning.MODERATE_PRUNE] = True
        elif marks['s'] >= 1:
            out[MarkMeaning.AGRESSIVE_PRUNE] = True
        
        if marks['t'] >= 0:
            out[MarkMeaning.RICH_FORMAT] = True
            if marks['t'] == 0:
                t = self._simple_time = time()
                out[MarkMeaning.RESET_TIMER] = t
            elif marks['t'] == 1:
                t = self._simple_time = time()
                out[MarkMeaning.START_TIMER] = t
            elif marks['t'] == 2:
                start, end = self._simple_time, time()
                self._simple_time = end
                out[MarkMeaning.STOP_TIMER] = (start, end)
        
        if marks['v'] >= 1:
            levels = ('trace', 'debug', 'info', 'warn', 'error', 'fatal')
            out[MarkMeaning.VERBOSITY] = levels[marks['v']]
        
        # postfix
        # if any((
        #     MarkMeaning.RESET_INDEX in out,
        #     MarkMeaning.EXPAND_MULTIPLE_LINES in out,
        # )):
        #     out[MarkMeaning.RICH_FORMAT] = True
        
        return out
